@@ -4,6 +4,10 @@ from __future__ import unicode_literals
 from __future__ import print_function
 import binascii
 import cgi
+try:
+    from html import escape as escape_html
+except ImportError:
+    from cgi import escape as escape_html
 from distutils.version import LooseVersion
 import glob
 import multiprocessing.pool
@@ -17,7 +21,7 @@ import socket
 import string
 import subprocess
 import sys
-import webbrowser
+from itertools import chain
 
 FISH_BIN_PATH = False  # will be set later
 IS_PY2 = sys.version_info[0] == 2
@@ -36,9 +40,20 @@ def isMacOS10_12_5_OrLater():
     version = platform.mac_ver()[0]
     return version and LooseVersion(version) >= LooseVersion('10.12.5')
 
+def is_wsl():
+    """ Return whether we are running under the Windows Subsystem for Linux """
+    if 'linux' in platform.system().lower():
+        with open('/proc/version', 'r') as f:
+            if 'Microsoft' in f.read():
+                return True
+    return False
+
 
 # Disable CLI web browsers
 term = os.environ.pop('TERM', None)
+# This import must be done with an empty $TERM, otherwise a command-line browser may be started
+# which will block the whole process - see https://docs.python.org/3/library/webbrowser.html
+import webbrowser
 if term:
     os.environ['TERM'] = term
 
@@ -123,7 +138,7 @@ def parse_color(color_str):
     comps = color_str.split(' ')
     color = 'normal'
     background_color = ''
-    bold, underline = False, False
+    bold, underline, italics, dim, reverse = False, False, False, False, False
     for comp in comps:
         # Remove quotes
         comp = comp.strip("'\" ")
@@ -131,6 +146,12 @@ def parse_color(color_str):
             bold = True
         elif comp == '--underline':
             underline = True
+        elif comp == '--italics':
+            italics = True
+        elif comp == '--dim':
+            dim = True
+        elif comp == '--reverse':
+            reverse = True
         elif comp.startswith('--background='):
             # Background color
             background_color = better_color(
@@ -140,7 +161,7 @@ def parse_color(color_str):
             color = better_color(color, parse_one_color(comp))
 
     return {"color": color, "background": background_color, "bold": bold,
-            "underline": underline}
+            "underline": underline, "italics": italics, "dim": dim, "reverse": reverse}
 
 
 def parse_bool(val):
@@ -234,7 +255,6 @@ def get_special_ansi_escapes():
 
 
 def append_html_for_ansi_escape(full_val, result, span_open):
-
     # Strip off the initial \x1b[ and terminating m
     val = full_val[2:-1]
 
@@ -251,10 +271,10 @@ def append_html_for_ansi_escape(full_val, result, span_open):
         result.append('<span style="color: ' + html_color + '">')
         return True  # span now open
 
-    # term8 foreground color
-    if val in [str(x) for x in range(30, 38)]:
+    # term16 foreground color
+    if val in (str(x) for x in chain(range(90, 97), range(30, 38))):
         close_span()
-        html_color = html_color_for_ansi_color_index(int(val) - 30)
+        html_color = html_color_for_ansi_color_index(int(val) - (30 if int(val) < 90 else 82))
         result.append('<span style="color: ' + html_color + '">')
         return True  # span now open
 
@@ -264,7 +284,7 @@ def append_html_for_ansi_escape(full_val, result, span_open):
         close_span()
         return False
 
-    # We don't handle bold or underline yet
+    # TODO We don't handle bold, underline, italics, dim, or reverse yet
 
     # Do nothing on failure
     return span_open
@@ -294,8 +314,9 @@ def ansi_to_html(val):
     reg = re.compile("""
         (                        # Capture
          \x1b                    # Escape
-         [^m]+                   # One or more non-'m's
+         [^m]*                   # Zero or more non-'m's
          m                       # Literal m terminates the sequence
+         \x0f?                   # HACK: A ctrl-o - this is how tmux' sgr0 ends
         )                        # End capture
         """, re.VERBOSE)
     separated = reg.split(val)
@@ -312,7 +333,7 @@ def ansi_to_html(val):
         if i % 2 == 0:
             # It's text, possibly empty
             # Clean up other ANSI junk
-            result.append(cgi.escape(strip_ansi(component)))
+            result.append(escape_html(strip_ansi(component)))
         else:
             # It's an escape sequence. Close the previous escape.
             span_open = append_html_for_ansi_escape(component, result,
@@ -563,6 +584,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
                          'param',
                          'comment',
                          'match',
+                         'selection',
                          'search_match',
                          'operator',
                          'escape',
@@ -570,6 +592,9 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
                          'redirection',
                          'valid_path',
                          'autosuggestion'
+                         'user',
+                         'host',
+                         'cancel'
                          ])
 
         # Here are our color descriptions
@@ -583,6 +608,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             'param': 'Command parameters',
             'comment': 'Comments start with #',
             'match': 'Matching parenthesis',
+            'selection': 'Selected text',
             'search_match': 'History searching',
             'history_current': 'Directory history',
             'operator': 'Like * and ~',
@@ -590,7 +616,10 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             'cwd': 'Current directory',
             'cwd_root': 'cwd for root user',
             'valid_path': 'Valid paths',
-            'autosuggestion': 'Suggested completion'
+            'autosuggestion': 'Suggested completion',
+            'user': 'Username in the prompt',
+            'host': 'Hostname in the prompt',
+            'cancel': 'The ^C cancel indicator'
         }
 
         out, err = run_fish_cmd('set -L')
@@ -675,6 +704,20 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         for line in out.split('\n'):
             comps = line.split(' ', 2)
 
+            # If we don't have "bind", a sequence and a mapping,
+            # it's not a valid binding.
+            if len(comps) < 3:
+                continue
+
+            # Store the "--preset" value for later
+            if comps[1] == '--preset':
+                preset = True
+                # There's possibly a way to do this faster, but it's not important.
+                comps = line.split(' ', 3)[1:]
+            elif comps[1] == '--user':
+                preset = False
+                comps = line.split(' ', 3)[1:]
+            # Check again if we removed the level.
             if len(comps) < 3:
                 continue
 
@@ -716,11 +759,17 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         return out
 
     def do_set_color_for_variable(self, name, color, background_color, bold,
-                                  underline):
+                                  underline, italics, dim, reverse):
+        "Sets a color for a fish color name, like 'autosuggestion'"
         if not color:
             color = 'normal'
-        "Sets a color for a fish color name, like 'autosuggestion'"
-        command = 'set -U fish_color_' + name
+        varname = 'fish_color_' + name
+        # If the name already starts with "fish_", use it as the varname
+        # This is needed for 'fish_pager_color' vars.
+        if name.startswith('fish_'):
+            varname = name
+        # TODO: Check if the varname is allowable.
+        command = 'set -U ' + varname
         if color:
             command += ' ' + color
         if background_color:
@@ -729,6 +778,12 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             command += ' --bold'
         if underline:
             command += ' --underline'
+        if italics:
+            command += ' --italics'
+        if dim:
+            command += ' --dim'
+        if reverse:
+            command += ' --reverse'
 
         out, err = run_fish_cmd(command)
         return out
@@ -739,7 +794,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
 
     def do_delete_history_item(self, history_item_text):
         # It's really lame that we always return success here
-        cmd = ('builtin history delete --exact -- %s; builtin history save' %
+        cmd = ('builtin history delete --case-sensitive --exact -- %s; builtin history save' %
                escape_fish_cmd(history_item_text))
         out, err = run_fish_cmd(cmd)
         return True
@@ -836,11 +891,15 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         return result
 
     def do_get_abbreviations(self):
-        out, err = run_fish_cmd('echo -n -s $fish_user_abbreviations\x1e')
-
-        lines = (x for x in out.rstrip().split('\x1e'))
-        abbrs = (re.split('[ =]', x, maxsplit=1) for x in lines if x)
-        result = [{'word': x, 'phrase': y} for x, y in abbrs]
+        # Example abbreviation line:
+        # abbr -a -U -- ls 'ls -a'
+        result = []
+        out, err = run_fish_cmd('abbr --show')
+        for line in out.rstrip().split('\n'):
+            if not line: continue
+            _, abbr = line.split(' -- ', 1)
+            word, phrase = abbr.split(' ', 1)
+            result.append({'word':word, 'phrase':phrase})
         return result
 
     def do_remove_abbreviation(self, abbreviation):
@@ -959,13 +1018,17 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             color = postvars.get('color')
             background_color = postvars.get('background_color')
             bold = postvars.get('bold')
+            italics = postvars.get('italics')
+            reverse = postvars.get('reverse')
+            dim = postvars.get('dim')
             underline = postvars.get('underline')
 
             if what:
                 # Not sure why we get lists here?
                 output = self.do_set_color_for_variable(
                     what[0], color[0], background_color[0],
-                    parse_bool(bold[0]), parse_bool(underline[0]))
+                    parse_bool(bold[0]), parse_bool(underline[0]), parse_bool(italics[0]),
+                    parse_bool(dim[0]), parse_bool(reverse[0]))
             else:
                 output = 'Bad request'
         elif p == '/get_function/':
@@ -1137,6 +1200,8 @@ fileurl = 'file://' + filename
 print("Web config started at '%s'. Hit enter to stop." % fileurl)
 if isMacOS10_12_5_OrLater():
     subprocess.check_call(['open', fileurl])
+elif is_wsl():
+    subprocess.call(['cmd.exe', '/c', "start %s" % url])
 else:
     webbrowser.open(fileurl)
 
