@@ -128,8 +128,9 @@ int builtin_complete(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     wcstring_list_t cmd_to_complete;
     wcstring_list_t path;
     wcstring_list_t wrap_targets;
+    bool preserve_order = false;
 
-    static const wchar_t *short_options = L":a:c:kp:s:l:o:d:frxeuAn:C::w:h";
+    static const wchar_t *const short_options = L":a:c:p:s:l:o:d:frxeuAn:C::w:hk";
     static const struct woption long_options[] = {{L"exclusive", no_argument, NULL, 'x'},
                                                   {L"no-files", no_argument, NULL, 'f'},
                                                   {L"require-parameter", no_argument, NULL, 'r'},
@@ -167,8 +168,7 @@ int builtin_complete(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                 break;
             }
             case 'k': {
-                // This is a no-op in fish 2.7. It is implemented in fish 3.0. We want it to be
-                // silently ignored if someone happens to use a completion that uses this flag.
+                preserve_order = true;
                 break;
             }
             case 'p':
@@ -318,8 +318,7 @@ int builtin_complete(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             recursion_level++;
 
             std::vector<completion_t> comp;
-            complete(do_complete_param, &comp, COMPLETION_REQUEST_DEFAULT,
-                     env_vars_snapshot_t::current());
+            complete(do_complete_param, &comp, COMPLETION_REQUEST_DEFAULT);
 
             for (size_t i = 0; i < comp.size(); i++) {
                 const completion_t &next = comp.at(i);
@@ -361,6 +360,9 @@ int builtin_complete(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         streams.out.append(complete_print());
     } else {
         int flags = COMPLETE_AUTO_SPACE;
+        if (preserve_order) {
+            flags |= COMPLETE_DONT_SORT;
+        }
 
         if (remove) {
             builtin_complete_remove(cmd_to_complete, path, short_opt.c_str(), gnu_opt, old_opt);
