@@ -2,6 +2,7 @@
 #include "config.h"  // IWYU pragma: keep
 
 #include <errno.h>
+#include <limits.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -11,7 +12,9 @@
 
 #include <algorithm>
 #include <memory>
+#include <numeric>
 #include <string>
+#include <vector>
 
 #include "builtin.h"
 #include "builtin_read.h"
@@ -19,7 +22,6 @@
 #include "complete.h"
 #include "env.h"
 #include "event.h"
-#include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "highlight.h"
 #include "history.h"
@@ -38,31 +40,41 @@ struct read_cmd_opts_t {
     const wchar_t *prompt_str = NULL;
     const wchar_t *right_prompt = L"";
     const wchar_t *commandline = L"";
+    // If a delimiter was given. Used to distinguish between the default
+    // empty string and a given empty delimiter.
+    bool have_delimiter = false;
+    wcstring delimiter;
     bool shell = false;
     bool array = false;
     bool silent = false;
     bool split_null = false;
+    bool to_stdout = false;
     int nchars = 0;
+    bool one_line = false;
 };
 
-static const wchar_t *short_options = L":ac:ghilm:n:p:suxzP:UR:";
-static const struct woption long_options[] = {{L"export", no_argument, NULL, 'x'},
-                                              {L"global", no_argument, NULL, 'g'},
-                                              {L"local", no_argument, NULL, 'l'},
-                                              {L"universal", no_argument, NULL, 'U'},
-                                              {L"unexport", no_argument, NULL, 'u'},
-                                              {L"prompt", required_argument, NULL, 'p'},
-                                              {L"prompt-str", required_argument, NULL, 'P'},
-                                              {L"right-prompt", required_argument, NULL, 'R'},
-                                              {L"command", required_argument, NULL, 'c'},
-                                              {L"mode-name", required_argument, NULL, 'm'},
-                                              {L"silent", no_argument, NULL, 'i'},
-                                              {L"nchars", required_argument, NULL, 'n'},
-                                              {L"shell", no_argument, NULL, 's'},
-                                              {L"array", no_argument, NULL, 'a'},
-                                              {L"null", no_argument, NULL, 'z'},
-                                              {L"help", no_argument, NULL, 'h'},
-                                              {NULL, 0, NULL, 0}};
+static const wchar_t *const short_options = L":ac:d:ghiLlm:n:p:suxzP:UR:LB";
+static const struct woption long_options[] = {
+    {L"array", no_argument, NULL, 'a'},
+    {L"command", required_argument, NULL, 'c'},
+    {L"delimiter", required_argument, NULL, 'd'},
+    {L"export", no_argument, NULL, 'x'},
+    {L"global", no_argument, NULL, 'g'},
+    {L"help", no_argument, NULL, 'h'},
+    {L"line", no_argument, NULL, 'L'},
+    {L"local", no_argument, NULL, 'l'},
+    {L"mode-name", required_argument, NULL, 'm'},
+    {L"nchars", required_argument, NULL, 'n'},
+    {L"null", no_argument, NULL, 'z'},
+    {L"prompt", required_argument, NULL, 'p'},
+    {L"prompt-str", required_argument, NULL, 'P'},
+    {L"right-prompt", required_argument, NULL, 'R'},
+    {L"shell", no_argument, NULL, 'S'},
+    {L"silent", no_argument, NULL, 's'},
+    {L"unexport", no_argument, NULL, 'u'},
+    {L"universal", no_argument, NULL, 'U'},
+    {NULL, 0, NULL, 0}
+};
 
 static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncss method)
                           int argc, wchar_t **argv, parser_t &parser, io_streams_t &streams) {
@@ -71,42 +83,40 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
     wgetopter_t w;
     while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
         switch (opt) {
-            case L'x': {
-                opts.place |= ENV_EXPORT;
+            case 'a': {
+                opts.array = true;
                 break;
             }
+            case L'c': {
+                opts.commandline = w.woptarg;
+                break;
+            }
+            case 'd': {
+                opts.have_delimiter = true;
+                opts.delimiter = w.woptarg;
+                break;
+            }
+            case 'i': {
+                streams.err.append_format(_(L"%ls: usage of -i for --silent is deprecated. Please use -s or --silent instead.\n"),
+                        cmd);
+                return STATUS_INVALID_ARGS;
+            }
             case L'g': {
                 opts.place |= ENV_GLOBAL;
                 break;
             }
+            case 'h': {
+                opts.print_help = true;
+                break;
+            }
+            case L'L': {
+                opts.one_line = true;
+                break;
+            }
             case L'l': {
                 opts.place |= ENV_LOCAL;
                 break;
             }
-            case L'U': {
-                opts.place |= ENV_UNIVERSAL;
-                break;
-            }
-            case L'u': {
-                opts.place |= ENV_UNEXPORT;
-                break;
-            }
-            case L'p': {
-                opts.prompt = w.woptarg;
-                break;
-            }
-            case L'P': {
-                opts.prompt_str = w.woptarg;
-                break;
-            }
-            case L'R': {
-                opts.right_prompt = w.woptarg;
-                break;
-            }
-            case L'c': {
-                opts.commandline = w.woptarg;
-                break;
-            }
             case L'm': {
                 streams.err.append_format(_(L"%ls: flags '--mode-name' / '-m' are now ignored. "
                                             L"Set fish_history instead.\n"),
@@ -130,26 +140,42 @@ static int parse_cmd_opts(read_cmd_opts_t &opts, int *optind,  //!OCLINT(high nc
                 }
                 break;
             }
+            case L'P': {
+                opts.prompt_str = w.woptarg;
+                break;
+            }
+            case L'p': {
+                opts.prompt = w.woptarg;
+                break;
+            }
+            case L'R': {
+                opts.right_prompt = w.woptarg;
+                break;
+            }
             case 's': {
+                opts.silent = true;
+                break;
+            }
+            case L'S': {
                 opts.shell = true;
                 break;
             }
-            case 'a': {
-                opts.array = true;
+            case L'U': {
+                opts.place |= ENV_UNIVERSAL;
                 break;
             }
-            case L'i': {
-                opts.silent = true;
+            case L'u': {
+                opts.place |= ENV_UNEXPORT;
+                break;
+            }
+            case L'x': {
+                opts.place |= ENV_EXPORT;
                 break;
             }
             case L'z': {
                 opts.split_null = true;
                 break;
             }
-            case 'h': {
-                opts.print_help = true;
-                break;
-            }
             case ':': {
                 builtin_missing_argument(parser, streams, cmd, argv[w.woptind - 1]);
                 return STATUS_INVALID_ARGS;
@@ -179,7 +205,7 @@ static int read_interactive(wcstring &buff, int nchars, bool shell, bool silent,
 
     wcstring read_history_ID = history_session_id();
     if (!read_history_ID.empty()) read_history_ID += L"_read";
-    reader_push(read_history_ID.c_str());
+    reader_push(read_history_ID);
 
     reader_set_left_prompt(prompt);
     reader_set_right_prompt(right_prompt);
@@ -320,26 +346,21 @@ static int read_one_char_at_a_time(int fd, wcstring &buff, int nchars, bool spli
     return exit_res;
 }
 
-/// The read builtin. Reads from stdin and stores the values in environment variables.
-int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
-    wchar_t *cmd = argv[0];
-    int argc = builtin_count_args(argv);
-    wcstring buff;
-    int exit_res = STATUS_CMD_OK;
-    read_cmd_opts_t opts;
-
-    int optind;
-    int retval = parse_cmd_opts(opts, &optind, argc, argv, parser, streams);
-    if (retval != STATUS_CMD_OK) return retval;
-
-    if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
-        return STATUS_CMD_OK;
+/// Validate the arguments given to `read` and provide defaults where needed.
+static int validate_read_args(const wchar_t *cmd, read_cmd_opts_t &opts, int argc,
+                              const wchar_t *const *argv, parser_t &parser, io_streams_t &streams) {
+    if (opts.prompt && opts.prompt_str) {
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"-p", L"-P");
+        builtin_print_help(parser, streams, cmd, streams.err);
+        return STATUS_INVALID_ARGS;
     }
 
-    if (opts.prompt && opts.prompt_str) {
-        streams.err.append_format(_(L"%ls: You can't specify both -p and -P\n"), cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+    if (opts.have_delimiter && opts.one_line) {
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"--delimiter", L"--line");
+        return STATUS_INVALID_ARGS;
+    }
+    if (opts.one_line && opts.split_null) {
+        streams.err.append_format(_(L"%ls: Options %ls and %ls cannot be used together\n"), cmd, L"-z", L"--line");
         return STATUS_INVALID_ARGS;
     }
 
@@ -364,15 +385,23 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         return STATUS_INVALID_ARGS;
     }
 
-    if (opts.array && optind + 1 != argc) {
-        streams.err.append_format(_(L"%ls: --array option requires a single variable name.\n"),
-                                  cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+    if (!opts.array && argc < 1 && !opts.to_stdout) {
+        streams.err.append_format(BUILTIN_ERR_MIN_ARG_COUNT1, cmd, 1, argc);
+        return STATUS_INVALID_ARGS;
+    }
+
+    if (opts.array && argc != 1) {
+        streams.err.append_format(BUILTIN_ERR_ARG_COUNT1, cmd, 1, argc);
+        return STATUS_INVALID_ARGS;
+    }
+
+    if (opts.to_stdout && argc > 0) {
+        streams.err.append_format(BUILTIN_ERR_MAX_ARG_COUNT1, cmd, 0, argc);
         return STATUS_INVALID_ARGS;
     }
 
     // Verify all variable names.
-    for (int i = optind; i < argc; i++) {
+    for (int i = 0; i < argc; i++) {
         if (!valid_var_name(argv[i])) {
             streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, argv[i]);
             builtin_print_help(parser, streams, cmd, streams.err);
@@ -380,80 +409,172 @@ int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         }
     }
 
-    // TODO: Determine if the original set of conditions for interactive reads should be reinstated:
-    // if (isatty(0) && streams.stdin_fd == STDIN_FILENO && !split_null) {
-    int stream_stdin_is_a_tty = isatty(streams.stdin_fd);
-    if (stream_stdin_is_a_tty && !opts.split_null) {
-        // We should read interactively using reader_readline(). This does not support splitting on
-        // null.
-        exit_res = read_interactive(buff, opts.nchars, opts.shell, opts.silent, opts.prompt,
-                                    opts.right_prompt, opts.commandline);
-    } else if (!opts.nchars && !stream_stdin_is_a_tty &&
-               lseek(streams.stdin_fd, 0, SEEK_CUR) != -1) {
-        exit_res = read_in_chunks(streams.stdin_fd, buff, opts.split_null);
-    } else {
-        exit_res = read_one_char_at_a_time(streams.stdin_fd, buff, opts.nchars, opts.split_null);
+    return STATUS_CMD_OK;
+}
+
+/// The read builtin. Reads from stdin and stores the values in environment variables.
+int builtin_read(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+    wchar_t *cmd = argv[0];
+    int argc = builtin_count_args(argv);
+    wcstring buff;
+    int exit_res = STATUS_CMD_OK;
+    read_cmd_opts_t opts;
+
+    int optind;
+    int retval = parse_cmd_opts(opts, &optind, argc, argv, parser, streams);
+    if (retval != STATUS_CMD_OK) return retval;
+    if (!opts.to_stdout) {
+        argc -= optind;
+        argv += optind;
     }
 
-    if (optind == argc || exit_res != STATUS_CMD_OK) {
-        // Define the var without any data. We do this because when this happens we want the user to
-        // be able to use the var but have it expand to nothing.
-        //
-        // TODO: For fish 3.0 we should mandate at least one var name.
-        if (argv[optind]) env_set(argv[optind], NULL, opts.place);
-        return exit_res;
+    if (argc == 0) {
+        opts.to_stdout = true;
     }
 
-    env_var_t ifs = env_get_string(L"IFS");
-    if (ifs.missing_or_empty()) {
-        // Every character is a separate token.
-        size_t bufflen = buff.size();
-        if (opts.array) {
-            if (bufflen > 0) {
-                wcstring chars(bufflen + (bufflen - 1), ARRAY_SEP);
-                wcstring::iterator out = chars.begin();
-                for (wcstring::const_iterator it = buff.begin(), end = buff.end(); it != end;
-                     ++it) {
-                    *out = *it;
-                    out += 2;
-                }
-                env_set(argv[optind], chars.c_str(), opts.place);
-            } else {
-                env_set(argv[optind], NULL, opts.place);
-            }
-        } else {  // not array
-            size_t j = 0;
-            for (; optind + 1 < argc; ++optind) {
-                if (j < bufflen) {
-                    wchar_t buffer[2] = {buff[j++], 0};
-                    env_set(argv[optind], buffer, opts.place);
+    if (opts.print_help) {
+        builtin_print_help(parser, streams, cmd, streams.out);
+        return STATUS_CMD_OK;
+    }
+
+    retval = validate_read_args(cmd, opts, argc, argv, parser, streams);
+    if (retval != STATUS_CMD_OK) return retval;
+
+    if (opts.one_line) {
+        // --line is the same as read -d \n repeated N times
+        opts.have_delimiter = true;
+        opts.delimiter = L"\n";
+        opts.split_null = false;
+        opts.shell = false;
+    }
+
+    wchar_t * const *var_ptr = argv;
+    auto vars_left = [&] () { return argv + argc - var_ptr; };
+    auto clear_remaining_vars = [&] () {
+        while (vars_left()) {
+            env_set_empty(*var_ptr, opts.place);
+            // env_set_one(*var_ptr, opts.place, L"");
+            ++var_ptr;
+        }
+    };
+
+    // Normally, we either consume a line of input or all available input. But if we are reading a line at
+    // a time, we need a middle ground where we only consume as many lines as we need to fill the given vars.
+    do {
+        buff.clear();
+
+        // TODO: Determine if the original set of conditions for interactive reads should be reinstated:
+        // if (isatty(0) && streams.stdin_fd == STDIN_FILENO && !split_null) {
+        int stream_stdin_is_a_tty = isatty(streams.stdin_fd);
+        if (stream_stdin_is_a_tty && !opts.split_null) {
+            // Read interactively using reader_readline(). This does not support splitting on null.
+            exit_res = read_interactive(buff, opts.nchars, opts.shell, opts.silent, opts.prompt,
+                                        opts.right_prompt, opts.commandline);
+        } else if (!opts.nchars && !stream_stdin_is_a_tty &&
+                   lseek(streams.stdin_fd, 0, SEEK_CUR) != -1) {
+            exit_res = read_in_chunks(streams.stdin_fd, buff, opts.split_null);
+        } else {
+            exit_res = read_one_char_at_a_time(streams.stdin_fd, buff, opts.nchars, opts.split_null);
+        }
+
+        if (exit_res != STATUS_CMD_OK) {
+            clear_remaining_vars();
+            return exit_res;
+        }
+
+        if (opts.to_stdout) {
+            streams.out.append(buff);
+            return exit_res;
+        }
+
+        if (!opts.have_delimiter) {
+            auto ifs = env_get(L"IFS");
+            if (!ifs.missing_or_empty()) opts.delimiter = ifs->as_string();
+        }
+
+        if (opts.delimiter.empty()) {
+            // Every character is a separate token with one wrinkle involving non-array mode where the
+            // final var gets the remaining characters as a single string.
+            size_t x = std::max(static_cast<size_t>(1), buff.size());
+            size_t n_splits = (opts.array || static_cast<size_t>(vars_left()) > x) ? x : vars_left();
+            wcstring_list_t chars;
+            chars.reserve(n_splits);
+            x = x - n_splits + 1;
+            int i = 0;
+            for (auto it = buff.begin(), end = buff.end(); it != end; ++i, ++it) {
+                if (opts.array || i + 1 < vars_left()) {
+                    chars.emplace_back(1, *it);
                 } else {
-                    env_set(argv[optind], L"", opts.place);
+                    chars.emplace_back(it, buff.end());
+                    break;
                 }
             }
-            if (optind < argc) env_set(argv[optind], &buff[j], opts.place);
-        }
-    } else if (opts.array) {
-        wcstring tokens;
-        tokens.reserve(buff.size());
-        bool empty = true;
 
-        for (wcstring_range loc = wcstring_tok(buff, ifs); loc.first != wcstring::npos;
-             loc = wcstring_tok(buff, ifs, loc)) {
-            if (!empty) tokens.push_back(ARRAY_SEP);
-            tokens.append(buff, loc.first, loc.second);
-            empty = false;
-        }
-        env_set(argv[optind], empty ? NULL : tokens.c_str(), opts.place);
-    } else {  // not array
-        wcstring_range loc = wcstring_range(0, 0);
+            if (opts.array) {
+                // Array mode: assign each char as a separate element of the sole var.
+                env_set(*var_ptr++, opts.place, chars);
+            } else {
+                // Not array mode: assign each char to a separate var with the remainder being assigned
+                // to the last var.
+                auto c = chars.begin();
+                for (; c != chars.end() && vars_left(); ++c) {
+                    env_set_one(*var_ptr++, opts.place, *c);
+                }
+            }
+        } else if (opts.array) {
+            // The user has requested the input be split into a sequence of tokens and all the tokens
+            // assigned to a single var. How we do the tokenizing depends on whether the user specified
+            // the delimiter string or we're using IFS.
+            if (!opts.have_delimiter) {
+                // We're using IFS, so tokenize the buffer using each IFS char. This is for backward
+                // compatibility with old versions of fish.
+                wcstring_list_t tokens;
 
-        while (optind < argc) {
-            loc = wcstring_tok(buff, (optind + 1 < argc) ? ifs : wcstring(), loc);
-            env_set(argv[optind], loc.first == wcstring::npos ? L"" : &buff.c_str()[loc.first],
-                    opts.place);
-            ++optind;
+                for (wcstring_range loc = wcstring_tok(buff, opts.delimiter);
+                     loc.first != wcstring::npos; loc = wcstring_tok(buff, opts.delimiter, loc)) {
+                    tokens.emplace_back(wcstring(buff, loc.first, loc.second));
+                }
+                env_set(*var_ptr++, opts.place, tokens);
+            } else {
+                // We're using a delimiter provided by the user so use the `string split` behavior.
+                wcstring_list_t splits;
+                split_about(buff.begin(), buff.end(), opts.delimiter.begin(), opts.delimiter.end(),
+                            &splits);
+
+                env_set(*var_ptr++, opts.place, splits);
+            }
+        } else {
+            // Not array mode. Split the input into tokens and assign each to the vars in sequence.
+            if (!opts.have_delimiter) {
+                // We're using IFS, so tokenize the buffer using each IFS char. This is for backward
+                // compatibility with old versions of fish.
+                wcstring_range loc = wcstring_range(0, 0);
+                while (vars_left()) {
+                    wcstring substr;
+                    loc = wcstring_tok(buff, (vars_left() > 1) ? opts.delimiter : wcstring(), loc);
+                    if (loc.first != wcstring::npos) {
+                        substr = wcstring(buff, loc.first, loc.second);
+                    }
+                    env_set_one(*var_ptr++, opts.place, substr);
+                }
+            } else {
+                // We're using a delimiter provided by the user so use the `string split` behavior.
+                wcstring_list_t splits;
+                // We're making at most argc - 1 splits so the last variable
+                // is set to the remaining string.
+                split_about(buff.begin(), buff.end(), opts.delimiter.begin(), opts.delimiter.end(),
+                            &splits, argc - 1);
+                assert(splits.size() <= (size_t) vars_left());
+                for (const auto &split : splits) {
+                    env_set_one(*var_ptr++, opts.place, split);
+                }
+            }
         }
+    } while (opts.one_line && vars_left());
+
+    if (!opts.array) {
+        // In case there were more args than splits
+        clear_remaining_vars();
     }
 
     return exit_res;
