@@ -44,6 +44,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include "fallback.h"  // IWYU pragma: keep
 #include "fish_version.h"
 #include "function.h"
+#include "future_feature_flags.h"
 #include "history.h"
 #include "io.h"
 #include "parser.h"
@@ -53,14 +54,11 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include "signal.h"
 #include "wutil.h"  // IWYU pragma: keep
 
-// PATH_MAX may not exist.
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
 // container to hold the options specified within the command line
 class fish_cmd_opts_t {
    public:
+    // Future feature flags values string
+    wcstring features;
     // Commands to be executed in place of interactive shell.
     std::vector<std::string> batch_cmds;
     // Commands to execute after the shell's config has been read.
@@ -91,44 +89,27 @@ extern "C" {
 int _NSGetExecutablePath(char *buf, uint32_t *bufsize);
 }
 
-/// Return the path to the current executable. This needs to be realpath'd.
-static std::string get_executable_path(const char *argv0) {
-    char buff[PATH_MAX];
-
-#if __APPLE__
-    // On OS X use it's proprietary API to get the path to the executable.
-    // This is basically grabbing exec_path after argc, argv, envp, ...: for us
-    // https://opensource.apple.com/source/adv_cmds/adv_cmds-163/ps/print.c
-    uint32_t buffSize = sizeof buff;
-    if (_NSGetExecutablePath(buff, &buffSize) == 0) return std::string(buff);
-#else
-    // On non-OS X UNIXes, try /proc directory.
-    ssize_t len;
-    len = readlink("/proc/self/exe", buff, sizeof buff - 1);  // Linux
-    if (len == -1) {
-        len = readlink("/proc/curproc/file", buff, sizeof buff - 1);  // BSD
-        if (len == -1) {
-            len = readlink("/proc/self/path/a.out", buff, sizeof buff - 1);  // Solaris
-        }
-    }
-    if (len > 0) {
-        buff[len] = '\0';
-        return std::string(buff);
-    }
-#endif
-
-    // Just return argv0, which probably won't work (i.e. it's not an absolute path or a path
-    // relative to the working directory, but instead something the caller found via $PATH). We'll
-    // eventually fall back to the compile time paths.
-    return std::string(argv0 ? argv0 : "");
-}
-
 static struct config_paths_t determine_config_directory_paths(const char *argv0) {
     struct config_paths_t paths;
     bool done = false;
     std::string exec_path = get_executable_path(argv0);
     if (get_realpath(exec_path)) {
-        debug(2, L"exec_path: '%s'", exec_path.c_str());
+        debug(2, L"exec_path: '%s', argv[0]: '%s'", exec_path.c_str(), argv0);
+        // TODO: we should determine program_name from argv0 somewhere in this file
+
+#ifdef CMAKE_BINARY_DIR
+        // Detect if we're running right out of the CMAKE build directory
+        if (string_prefixes_string(CMAKE_BINARY_DIR, exec_path.c_str())) {
+            debug(2, "Running out of build directory, using paths relative to CMAKE_SOURCE_DIR:\n %s", CMAKE_SOURCE_DIR);
+
+            done = true;
+            paths.data = wcstring{L"" CMAKE_SOURCE_DIR} + L"/share";
+            paths.sysconf = wcstring{L"" CMAKE_SOURCE_DIR} + L"/etc";
+            paths.doc = wcstring{L"" CMAKE_SOURCE_DIR} + L"/user_doc/html";
+            paths.bin = wcstring{L"" CMAKE_BINARY_DIR};
+        }
+#endif
+
         if (!done) {
             // The next check is that we are in a reloctable directory tree
             const char *installed_suffix = "/bin/fish";
@@ -238,15 +219,17 @@ int run_command_list(std::vector<std::string> *cmds, const io_chain_t &io) {
 
 /// Parse the argument list, return the index of the first non-flag arguments.
 static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
-    static const char *short_opts = "+hilnvc:C:p:d:D:";
+    static const char * const short_opts = "+hPilnvc:C:p:d:f:D:";
     static const struct option long_opts[] = {{"command", required_argument, NULL, 'c'},
                                               {"init-command", required_argument, NULL, 'C'},
+                                              {"features", required_argument, NULL, 'f'},
                                               {"debug-level", required_argument, NULL, 'd'},
                                               {"debug-stack-frames", required_argument, NULL, 'D'},
                                               {"interactive", no_argument, NULL, 'i'},
                                               {"login", no_argument, NULL, 'l'},
                                               {"no-execute", no_argument, NULL, 'n'},
                                               {"profile", required_argument, NULL, 'p'},
+                                              {"private", no_argument, NULL, 'P'},
                                               {"help", no_argument, NULL, 'h'},
                                               {"version", no_argument, NULL, 'v'},
                                               {NULL, 0, NULL, 0}};
@@ -277,6 +260,10 @@ static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
                 }
                 break;
             }
+            case 'f': {
+                opts->features = str2wcstring(optarg);
+                break;
+            }
             case 'h': {
                 opts->batch_cmds.push_back("__fish_print_help fish");
                 break;
@@ -298,6 +285,10 @@ static int fish_parse_opt(int argc, char **argv, fish_cmd_opts_t *opts) {
                 g_profiling_active = true;
                 break;
             }
+            case 'P': {
+                start_private_mode();
+                break;
+            }
             case 'v': {
                 fwprintf(stdout, _(L"%s, version %s\n"), PACKAGE_NAME, get_fish_version());
                 exit(0);
@@ -353,8 +344,8 @@ int main(int argc, char **argv) {
     // struct stat tmp;
     // stat("----------FISH_HIT_MAIN----------", &tmp);
 
+    const char *dummy_argv[2] = {"fish", NULL};
     if (!argv[0]) {
-        static const char *dummy_argv[2] = {"fish", NULL};
         argv = (char **)dummy_argv;  //!OCLINT(parameter reassignment)
         argc = 1;                    //!OCLINT(parameter reassignment)
     }
@@ -375,25 +366,24 @@ int main(int argc, char **argv) {
 
     const struct config_paths_t paths = determine_config_directory_paths(argv[0]);
     env_init(&paths);
+    // Set features early in case other initialization depends on them.
+    // Start with the ones set in the environment, then those set on the command line (so the
+    // command line takes precedence).
+    if (auto features_var = env_get(L"fish_features")) {
+        for (const wcstring &s : features_var->as_list()) {
+            mutable_fish_features().set_from_string(s);
+        }
+    }
+    mutable_fish_features().set_from_string(opts.features);
     proc_init();
-    event_init();
     builtin_init();
-    function_init();
     misc_init();
     reader_init();
-    history_init();
 
     parser_t &parser = parser_t::principal_parser();
 
     const io_chain_t empty_ios;
     if (read_init(paths)) {
-        // TODO: Remove this once we're confident that not blocking/unblocking every signal around
-        // some critical sections is no longer necessary.
-        env_var_t fish_no_signal_block = env_get_string(L"FISH_NO_SIGNAL_BLOCK");
-        if (!fish_no_signal_block.missing_or_empty() && !from_string<bool>(fish_no_signal_block)) {
-            ignore_signal_block = false;
-        }
-
         // Stomp the exit status of any initialization commands (issue #635).
         proc_set_last_status(STATUS_CMD_OK);
 
@@ -427,8 +417,7 @@ int main(int argc, char **argv) {
                 for (char **ptr = argv + my_optind; *ptr; ptr++) {
                     list.push_back(str2wcstring(*ptr));
                 }
-                auto val = list_to_array_val(list);
-                env_set(L"argv", *val == ENV_NULL ? NULL : val->c_str(), 0);
+                env_set(L"argv", ENV_DEFAULT, list);
 
                 const wcstring rel_filename = str2wcstring(file);
 
@@ -448,7 +437,10 @@ int main(int argc, char **argv) {
 
     int exit_status = res ? STATUS_CMD_UNKNOWN : proc_get_last_status();
 
+    // TODO: The generic process-exit event is useless and unused.
+    // Remove this in future.
     proc_fire_event(L"PROCESS_EXIT", EVENT_EXIT, getpid(), exit_status);
+    event_fire_generic(L"fish_exit");
 
     restore_term_mode();
     restore_term_foreground_process_group();
@@ -457,11 +449,8 @@ int main(int argc, char **argv) {
         parser.emit_profiling(s_profiling_output_filename);
     }
 
-    history_destroy();
+    history_save_all();
     proc_destroy();
-    builtin_destroy();
-    reader_destroy();
-    event_destroy();
     exit_without_destructors(exit_status);
     return EXIT_FAILURE;  // above line should always exit
 }
