@@ -5,7 +5,12 @@
 #include "fish_version.h"
 
 #ifndef FISH_BUILD_VERSION
-#include "fish-build-version.h"
+// The contents of FISH-BUILD-VERSION-FILE looks like:
+// FISH_BUILD_VERSION="2.7.1-62-gc0480092-dirty"
+// Arrange for it to become a variable.
+static const char * const
+#include "FISH-BUILD-VERSION-FILE"
+    ;
 #endif
 
 /// Return fish shell version.
