@@ -14,11 +14,7 @@ enum {
     // delivered, or when an exception happened.
     R_NULL = R_MIN,
     R_EOF,
-    // Key codes for inputrc-style keyboard functions that are passed on to the caller of
-    // input_read().
-    //
-    // NOTE: If you modify this sequence of symbols you must update the name_arr, code_arr and
-    // desc_arr variables in input.cpp to match!
+
     R_BEGINNING_OF_LINE,
     R_END_OF_LINE,
     R_FORWARD_CHAR,
@@ -36,6 +32,7 @@ enum {
     R_YANK_POP,
     R_COMPLETE,
     R_COMPLETE_AND_SEARCH,
+    R_PAGER_TOGGLE_SEARCH,
     R_BEGINNING_OF_HISTORY,
     R_END_OF_HISTORY,
     R_BACKWARD_KILL_LINE,
@@ -70,13 +67,19 @@ enum {
     R_KILL_SELECTION,
     R_FORWARD_JUMP,
     R_BACKWARD_JUMP,
+    R_FORWARD_JUMP_TILL,
+    R_BACKWARD_JUMP_TILL,
     R_AND,
     R_CANCEL,
+    R_REPEAT_JUMP,
+    R_REVERSE_REPEAT_JUMP,
+
     R_TIMEOUT,  // we didn't get interactive input within wait_on_escape_ms
-    R_MAX = R_CANCEL,
-    // This is a special psuedo-char that is not used other than to mark the end of the the special
-    // characters so we can sanity check the enum range.
-    R_SENTINAL
+
+    // The range of key codes for inputrc-style keyboard functions that are passed on to the caller
+    // of input_read().
+    R_BEGIN_INPUT_FUNCTIONS = R_BEGINNING_OF_LINE,
+    R_END_INPUT_FUNCTIONS = R_REVERSE_REPEAT_JUMP + 1
 };
 
 /// Init the library.
