@@ -30,10 +30,10 @@ void signal_handle(int sig, int do_handle);
 void signal_unblock_all();
 
 /// Block all signals.
-void signal_block(bool force = false);
+void signal_block();
 
 /// Unblock all signals.
-void signal_unblock(bool force = false);
+void signal_unblock();
 
 /// Returns true if signals are being blocked.
 bool signal_is_blocked();
@@ -41,5 +41,16 @@ bool signal_is_blocked();
 /// Returns signals with non-default handlers.
 void get_signals_with_handlers(sigset_t *set);
 
-extern bool ignore_signal_block;
+/// A RAII wrapper for signal_block/signal_unblock that triggers a signal block on creation, and then
+/// automatically releases the block when the object is destroyed, handling control flow and exceptions.
+struct signal_block_t {
+    signal_block_t() {
+        signal_block();
+    }
+
+    ~signal_block_t() {
+        signal_unblock();
+    }
+};
+
 #endif
